﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Diagnostics;

using MathNet.Numerics;
using MathNet.Numerics.LinearAlgebra.Double;


//Kinect headers
using Microsoft.Kinect;
using Microsoft.Speech.AudioFormat;
using Microsoft.Speech.Recognition;


//OpenTK headers
using OpenTK;
using OpenTK.Graphics;
using OpenTK.Graphics.OpenGL;
using System.Threading;

namespace Final_Demo
{
    public partial class Form1 : Form
    {

        //Result List
        List<float> ghistResultConfidence; // Distance of Histogram - lower means closer(similar)
        List<int> ghistResultIndex; // Order of Results
        //List<float> lscResultConfidence; // Same for LSC.
        //List<int> lscResultIndex;

        static StreamWriter writeToFile; //write points to file

        static int NUMBER_OF_IMAGES = 10; // images to show

        DenseMatrix x = DenseMatrix.OfArray(new double[6, 1] { { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 } });
        // initial uncertainty
        DenseMatrix P = DenseMatrix.OfArray(new double[6, 6] { { 1000, 0, 0, 0, 0, 0 }, { 0, 1000, 0, 0, 0, 0 }, { 0, 0, 1000, 0, 0, 0 }, { 0, 0, 0, 1000, 0, 0 }, { 0, 0, 0, 0, 1000, 0 }, { 0, 0, 0, 0, 0, 1000 } });

        #region Kinect Variables

        //Speech Recognition
        SpeechRecognitionEngine speechEngine;

        bool IsRegisteringHandMovement = false;

        //Variables to set Elbow Above Gesture - To start
        float SAFE_DISTANCE_BW_HAND_ELBOW_ABOVE = 0.05f;
        const int TotalNumberOfVerticesToEstablishHandStableAboveElbow = 10;
        int CurrentVerticesEstablishedAboveElbow = 0;

        //Variables to set Elbow Below Gesture - To reset
        float SAFE_DISTANCE_BW_HAND_ELBOW_BELOW = 0.2f;
        const int TotalNumberOfVerticesToEstablishReset = 50;
        int CurrentVerticesEstablishedBelowElbow = 0;

        bool IsPaused = false;

        bool IsHearingOn = true;

        //const float THRESHOLD_LEVEL_NEAREST_POINTS = 0.0001f; //0.01 * 0.01
        const float THRESHOLD_POINTS_INTERPOLATION = 100.01f; //0.01 //so, this is the basic unit above which new points are interpolated as needed
        const float THRESHOLD_MINIMUM_DISTANCE_TO_QUALIFY_LONG_DISTANCE_ERROR = 0.3f; //0.3f

        static bool IsRightHandDrawing = true;

        static private DepthImagePixel[] _DepthPixels;

        static int x_right_hand, y_right_hand, depth_right_hand;

        static int WIDTH_IMAGE = 640; //640
        static int HEIGHT_IMAGE = 480; //480
        static int NORMALIZE_DEPTH = 1200; //maximum depth for near mode (3000)

        static int pixelIndex; //declared here so not created in every draw loop

        static private KinectSensor sensor;

        const int DEPTH_THRESHOLD_FINGER_HAND = 80; //if problem, reduce it

        #endregion

        #region OpenGL(OpenTK) Related Variables

        bool IsGLControlLoaded = false;


        //static string Input_Model_Filename = "hand map.obj";

        //static string CommonLocation = @"C:\Users\a_z65\Desktop\Final Demo\";

        static string FileWriteVertices = "Vertices.obj";
        static string ThumbNailLocation = "Thumbnails\\";
        static string FileToHistogram = "ThreshVertices.obj";

        static string CppExeLocation = "preProcessing.exe";
        static string CppArguments = String.Format("\"{0}\" \"{1}\"", "Vertices.obj", "PCAVertices.off");

        // Global-Histogram-specific strings for executing ghist.jar
        // Integrated by Mario on 7/28/2014 at 3:00 pm
        static string JAVA_EXEC = "java.exe";
        static string GHIST_JAR = "ghist.jar";
        static string GHIST_DIR = "ghists";
        static string GHIST_OUT = "ghist_results.txt";
        static string GHIST_ARG = String.Format("-jar \"{0}\" \"{1}\" \"{2}\" \"{3}\" --vanilla", GHIST_JAR, GHIST_DIR, FileToHistogram, GHIST_OUT);

        // Global-Histogram-specific strings for executing ghist.jar
        // Integrated by Mario on 7/28/2014 at 3:00 pm
        //static string LSC_EXEC = "SHREC3dSketchRetrievalOnline.exe";
        //static string LSC_DIR = "lschists";
        //static string LSC_OUT = "lsc_results.txt";
        //static string LSC_ARG = String.Format("{0} {1} {2}", LSC_DIR, FileToHistogram, LSC_OUT);

        //Projection
        const int Projection_Ortho = 1;
        const int Projcetion_Perspective = 2;
        static int project_mode = 0;

        //Box
        static bool draw_bounding_box = false;

        //Resize the window
        double aspect_ratio;

        //Draw xy-plane and axis
        static float centerX = 0, centerY = 0, centerZ = 0;
        static bool draw_xy_plane = true;
        static bool draw_axis = true;

        //Transform
        enum Transform { TRANSFORM_NONE, TRANSFORM_ROTATE, TRANSFORM_TRANSLATE, TRANSFORM_SCALE };

        static int xform_mode = (int)Transform.TRANSFORM_NONE;
        static int press_x, press_y;
        static float x_angle = 0.0f;
        static float y_angle = 0.0f;
        static float z_angle = 0.0f;
        static float scale_size = 2.0f;
        static float x_size = -0.5f;
        static float y_size = -0.4f;


        //Display
        enum Modes { OBJ_POINT, OBJ_LINE };
        static int current_object_mode = (int)Modes.OBJ_POINT;




        //[StructLayout(LayoutKind.Sequential)]
        struct Vertex
        { // mimic InterleavedArrayFormat.T2fN3fV3f
            public Vector3 Normal;
            public Vector3 Position;
        }

        struct Face                                     //----------------------Correct??
        { // mimic InterleavedArrayFormat.T2fN3fV3f
            public int v1, v2, v3;
        }


        //Data Structure of Model
        static List<Vertex> vertices;
        static List<Face> faces;

        #endregion


        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            TxtTotalResults.Text = "First " + NUMBER_OF_IMAGES.ToString();

            #region Kinect Initialization

            foreach (var potentialSensor in KinectSensor.KinectSensors)
            {
                if (potentialSensor.Status == KinectStatus.Connected)
                {
                    sensor = potentialSensor;
                    break;
                }
            }

            if (sensor != null)
            {
                // Turn on the skeleton stream to receive skeleton frames

                TransformSmoothParameters smoothParams = new TransformSmoothParameters();
                smoothParams.JitterRadius = 1.0f;
                //smoothParams.Correction = 0.5f;
                //smoothParams.MaxDeviationRadius = 0.05f;
                smoothParams.Prediction = 0.05f;
                smoothParams.Smoothing = 0.04f;

                //Small -> {0.5f, 0.5f, 0.5f, 0.05f, 0.04f};
                //Medium -> {0.5f, 0.1f, 0.5f, 0.1f, 0.1f};
                //Large -> {0.7f, 0.3f, 1.0f, 1.0f, 1.0f};



                sensor.SkeletonStream.Enable(smoothParams);
                sensor.DepthStream.Enable(DepthImageFormat.Resolution640x480Fps30);

                DepthImageStream depthStream = sensor.DepthStream;
                _DepthPixels = new DepthImagePixel[depthStream.FramePixelDataLength];

                sensor.AllFramesReady += AllFramesReady;

                sensor.Start();
            }//if sensor is not null, ie, some sensor attached

            if (null == sensor)
            {
                MessageBox.Show("Kinect Not Connected!");
            }


            #endregion

            if (sensor != null)
            {
                #region Initialize Voice Recognition

                RecognizerInfo recognizer_info = GetKinectRecognizer();

                if (recognizer_info != null)
                {

                    speechEngine = new SpeechRecognitionEngine(recognizer_info.Id);

                    //Creating Grammer Programmatically

                    Choices MyGrammar = new Choices();


                    //This is the vocabulary of the program.
                    //SemanticResultValue("X", "Y")
                    //X: is the word/phrase the person is supposed to speak
                    //Y: is the category the words correspond to.
                    //Using this system, you can speak different phrases
                    //like "reset", "erase", "erase all", "start again" and they will do the same thing
                    #region My Vocabulary

                    MyGrammar.Add(new SemanticResultValue("start", "Start"));

                    MyGrammar.Add(new SemanticResultValue("draw with right hand", "Right"));
                    MyGrammar.Add(new SemanticResultValue("use right hand", "Right"));
                    MyGrammar.Add(new SemanticResultValue("right hand", "Right"));
                    MyGrammar.Add(new SemanticResultValue("right", "Right"));

                    MyGrammar.Add(new SemanticResultValue("draw with left hand", "Left"));
                    MyGrammar.Add(new SemanticResultValue("use left hand", "Left"));
                    MyGrammar.Add(new SemanticResultValue("left hand", "Left"));
                    MyGrammar.Add(new SemanticResultValue("left", "Left"));


                    MyGrammar.Add(new SemanticResultValue("show more results", "ShowMoreResults"));

                    MyGrammar.Add(new SemanticResultValue("pause", "Pause"));
                    MyGrammar.Add(new SemanticResultValue("unpause", "Pause"));
                    MyGrammar.Add(new SemanticResultValue("resume", "Pause"));


                    MyGrammar.Add(new SemanticResultValue("switch hands", "Switch"));

                    MyGrammar.Add(new SemanticResultValue("reset", "Reset"));
                    MyGrammar.Add(new SemanticResultValue("start over", "Reset"));
                    MyGrammar.Add(new SemanticResultValue("erase", "Reset"));
                    MyGrammar.Add(new SemanticResultValue("do it again", "Reset"));
                    MyGrammar.Add(new SemanticResultValue("restart", "Reset"));

                    MyGrammar.Add(new SemanticResultValue("reset perspective", "Perspective"));
                    MyGrammar.Add(new SemanticResultValue("front view", "Perspective"));

                    MyGrammar.Add(new SemanticResultValue("side view", "SideView"));


                    MyGrammar.Add(new SemanticResultValue("search", "Search"));
                    MyGrammar.Add(new SemanticResultValue("finished", "Search"));
                    MyGrammar.Add(new SemanticResultValue("done", "Search"));


                    MyGrammar.Add(new SemanticResultValue("point mode", "PointMode"));
                    MyGrammar.Add(new SemanticResultValue("line mode", "LineMode"));

                    MyGrammar.Add(new SemanticResultValue("save", "Save"));


                    /*
                    MyGrammar.Add(new SemanticResultValue("exit program", "Exit"));
                    MyGrammar.Add(new SemanticResultValue("close program", "Exit"));

                    for (int i = 1; i <= 100; i++)
                    {
                        MyGrammar.Add(new SemanticResultValue(Counting[i - 1], Counting[i - 1]));
                        MyGrammar.Add(new SemanticResultValue("show " + Counting[i - 1] + "results", Counting[i - 1]));

                    }
                     * 
                     */

                    #endregion

                    GrammarBuilder grammar_builder = new GrammarBuilder { Culture = recognizer_info.Culture };

                    grammar_builder.Append(MyGrammar);

                    Grammar grammar = new Grammar(grammar_builder);

                    speechEngine.LoadGrammar(grammar);

                    speechEngine.SpeechRecognized += SpeechRecognized;

                    speechEngine.SpeechRecognitionRejected += SpeechRejected;

                    speechEngine.SetInputToAudioStream(sensor.AudioSource.Start(), new SpeechAudioFormatInfo(EncodingFormat.Pcm, 16000, 16, 1, 32000, 2, null));

                    speechEngine.RecognizeAsync(RecognizeMode.Multiple);

                }//if recognizer_info is not null

                else
                {

                    TxtVoiceCommand.Text = "Cannot Intialize Voice Recognition";

                }//if recognizer_info is NULL

                #endregion
            }

            ghistResultConfidence = new List<float>();
            ghistResultIndex = new List<int>();
            //lscResultConfidence = new List<float>();
            //lscResultIndex = new List<int>();

            vertices = new List<Vertex>();
            faces = new List<Face>();

            writeToFile = new StreamWriter(FileWriteVertices, false);
        }

        private void MyGlControl_Load(object sender, EventArgs e)
        {
            IsGLControlLoaded = true;
            GL.ClearColor(Color.BlueViolet);

            Application.Idle += Application_Idle;

            GL.ClearColor(Color.White);
            GL.Enable(EnableCap.DepthTest);
        }

        // application idle event
        void Application_Idle(object sender, EventArgs e)
        {
            //Use if needed!
            //fps setting here with the help of global Stopwatch()

            MyGlControl.Invalidate();
        }

        // resize MyGlControl
        private void MyGlControl_Resize(object sender, EventArgs e)
        {
            GL.Viewport(0, 0, Width, Height);

            aspect_ratio = Width / (double)Height;

            OpenTK.Matrix4 perspective = OpenTK.Matrix4.CreatePerspectiveFieldOfView(MathHelper.PiOver4, (float)aspect_ratio, 1, 64);
            GL.MatrixMode(MatrixMode.Projection);
            GL.LoadMatrix(ref perspective);

            MyGlControl.Invalidate(); //so, no matter what, it updates the MyGlControl
        }


        private void MyGlControl_Paint(object sender, PaintEventArgs e)
        {
            if (!IsGLControlLoaded) // Play nice
                return;

            GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);

            OpenTK.Matrix4 lookat = OpenTK.Matrix4.LookAt(0, 5, 5, 0, 0, 0, 0, 1, 0);
            GL.MatrixMode(MatrixMode.Modelview);
            GL.LoadMatrix(ref lookat);

            rendering_routine();

            MyGlControl.SwapBuffers();  //swapping buffers 

            //System.Threading.Thread.Sleep(1);// [Is it really needed in form application]
        }


        //If you want to Load a model, instead of drawing it, press the "1" button and set the 
        //"Input_Model_Filename" value to the filename of the .obj model you want to load
        #region Load Model

        //Load a Model
        /*
        static void LoadModel()
        {


            StreamReader reader = new StreamReader(Input_Model_Filename);
            string Line = "";
            string[] separated_numbers; //0-> v or f 

            while (true)
            {
                Line = reader.ReadLine();

                if (Line == null || Line == "")
                    break;

                if (Line[0] == 'v')
                {

                    separated_numbers = Line.Split(' ');


                    //Line[0]=v
                    Vertex vertex = new Vertex();
                    vertex.Position.X = float.Parse(separated_numbers[1]);
                    vertex.Position.Y = float.Parse(separated_numbers[2]);
                    vertex.Position.Z = float.Parse(separated_numbers[3]);

                    vertex.Normal.X = 0; vertex.Normal.Y = 0; vertex.Normal.Z = 0;
                    vertices.Add(vertex);

                    //Console.WriteLine(vertex.Position.X.ToString() + ", " + vertex.Position.Y.ToString() + ", " + vertex.Position.Z.ToString());


                }//its a vertex

                else if (Line[0] == 'f')
                {

                    
                    separated_numbers = Line.Split(' ');
                    
                    //Line[0]=v
                    Face face = new Face();
                    face.v1 = Int32.Parse(separated_numbers[1]) - 1;
                    face.v2 = Int32.Parse(separated_numbers[2]) - 1;
                    face.v3 = Int32.Parse(separated_numbers[3]) - 1;// the index in M is 1-based!
                    faces.Add(face);
                    



                }//its a face


                else
                {
                    //Normal and then not important
                }




            }//while reading file


            reader.Close();

            Console.WriteLine("Read File");

        }
        */

        #endregion


        #region OpenTK Function Routines

        /*-------------------------------------------------------------------------
          render all the points or lines 
          ---------------------------------------------------------------------------*/
        void rendering_routine()
        {
            GL.Enable(EnableCap.DepthTest);
            // Just clean the screen
            GL.ClearColor(Color.White);

            // lighting
            float[] light_position = { 1.0f, 1.0f, 1.0f, 0.0f };
            float[] white_light = { 1.0f, 1.0f, 1.0f, 1.0f };
            float[] lmodel_ambient = { 1.0f, 0.1f, 0.1f, 1.0f };

            GL.Light(LightName.Light0, LightParameter.Position, light_position);
            GL.Light(LightName.Light0, LightParameter.Diffuse, white_light);
            GL.Light(LightName.Light0, LightParameter.Specular, white_light);


            GL.LightModel(LightModelParameter.LightModelAmbient, lmodel_ambient);
            GL.Enable(EnableCap.Lighting);
            GL.Enable(EnableCap.Light0);

            // setup the perspective projection
            GL.MatrixMode(MatrixMode.Projection);
            GL.LoadIdentity();

            if (Projection_Ortho == project_mode)
            {
                GL.Ortho(-2 * aspect_ratio, 2 * aspect_ratio, -2, 2, 0.1, 100);
            }
            
            // not orthogonal projection
            else
            {
                // gluPerspective(60, AR, .1, 1000.0); //-------------------------------------------------------- 
                //OpenTK.Matrix4.Perspective(60, AR, .1f, 1000.0f);
                //OpenTK.Matrix4 perspective = OpenTK.Matrix4.CreatePerspectiveFieldOfView(60, AR, .1f, 1000.0f); //-----Didn't use the return OpenTK.Matrix4 
                OpenTK.Matrix4 perspective = OpenTK.Matrix4.CreatePerspectiveFieldOfView(1.047198f, (float)aspect_ratio, .1f, 1000.0f); //-----Didn't use the return OpenTK.Matrix4 

                OpenTK.Matrix4 lookat = OpenTK.Matrix4.LookAt(0, 0, 5, 0, 0, 0, 0, 1, 0); //----Didn't use the return OpenTK.Matrix4
                //setup the camera

                //gluLookAt(0, 0, 5, 0, 0, 0, 0, 1, 0); //---------------------------------------------------------

                GL.MatrixMode(MatrixMode.Projection);
                GL.LoadIdentity();
                GL.MultMatrix(ref perspective);

                GL.MatrixMode(MatrixMode.Modelview);
                GL.LoadIdentity();
                GL.MultMatrix(ref lookat);
            }



            // rotate and scale the object
            GL.Rotate(x_angle, 0, 1, 0);
            GL.Rotate(y_angle, 1, 0, 0);

            GL.Scale(scale_size, scale_size, scale_size);
            GL.Translate(x_size, y_size, 0);


            if (current_object_mode == (int)Modes.OBJ_POINT)
            {
                GL.Disable(EnableCap.Lighting);
                //Points
                GL.Color3(1.0f, 0.0f, 0.0f);   //conversion from 1 -> 255

                GL.PointSize(4f);
                GL.Begin(BeginMode.Points);
                for (int i = 0; i < vertices.Count; i++)
                {
                    Vertex V = vertices[i];
                    GL.Vertex3(V.Position.X, V.Position.Y, V.Position.Z);
                }
                GL.End();

                /* FingerTip
                GL.Color3(0.0f, 1.0f, 0.0f);
                GL.PointSize(5f);
                GL.Begin(BeginMode.Points);
                GL.Vertex3(0.1984375, 0.1583333, 1.175);
                GL.End();
                 */

                GL.Enable(EnableCap.Lighting);
            }


            else if (current_object_mode == (int)Modes.OBJ_LINE)
            {
                GL.Disable(EnableCap.Lighting);


                GL.LineWidth(3.0f);
                GL.Color3(0, 0.5, 1);


                for (int i = 0; i < vertices.Count - 1; i++)
                {
                    GL.Begin(BeginMode.Lines);
                    Vertex V = vertices[i];
                    Vertex V2 = vertices[i + 1];
                    GL.Vertex3(V.Position.X, V.Position.Y, V.Position.Z);
                    GL.Vertex3(V2.Position.X, V2.Position.Y, V2.Position.Z);
                    GL.End();
                }
                GL.Enable(EnableCap.Lighting);
            }

            // if it's not registering a hand movement then display a point
            // to show the current location of the hand on the screen
            if (IsRegisteringHandMovement == false)
            {
                GL.Disable(EnableCap.Lighting);
                //Points
                GL.Color3(0.0f, 1.0f, 0.0f);   //conversion from 1 -> 255

                GL.PointSize(10f);
                GL.Begin(BeginMode.Points);
                GL.Vertex3((float)x_right_hand / WIDTH_IMAGE, 1.0f - (float)y_right_hand / HEIGHT_IMAGE, (float)depth_right_hand / NORMALIZE_DEPTH);
                GL.End();

                GL.Enable(EnableCap.Lighting);
            } 


            GL.Disable(EnableCap.Lighting);

            //draw bounding box
            if (draw_bounding_box == true)
                drawboundingbox();

            // draw xy_plane and axis  
            if (draw_xy_plane == true) { displayfloor(); }
            if (draw_axis == true) { displayaxes(); }

            GL.Enable(EnableCap.Lighting);

            //swap the buffer->Swapped in Render function
            //glutSwapBuffers();
            //glutPostRedisplay();
        }

        // perform the kalman filter based on a measurement
        // in 3D space
        Vertex performKalman(Vertex measurement)
        {
            return kalmanFilter(x, P, measurement);
        }

        // helper function for the kalman filter
        // returns the smoothed vertex (x, y, z) after applying the filter
        Vertex kalmanFilter(DenseMatrix x, DenseMatrix P, Vertex measurement)
        {
            DenseMatrix F = DenseMatrix.OfArray(new double[6, 6] { { 1, 0, 0, 1, 0, 0 }, { 0, 1, 0, 0, 1, 0 }, { 0, 0, 1, 0, 0, 1 }, { 0, 0, 0, 1, 0, 0 }, { 0, 0, 0, 0, 1, 0 }, { 0, 0, 0, 0, 0, 1 } });
            DenseMatrix H = DenseMatrix.OfArray(new double[3, 6] { { 1, 0, 0, 0, 0, 0 }, { 0, 1, 0, 0, 0, 0 }, { 0, 0, 1, 0, 0, 0 } });
            DenseMatrix R = DenseMatrix.OfArray(new double[3, 3] { { 100, 0, 0 }, { 0, 100, 0 }, { 0, 0, 100 } });
            Matrix I = DenseMatrix.Identity(6);

            Vertex v = measurement;
            double[,] pts = new double[3, 1] { { v.Position.X }, { v.Position.Y }, { v.Position.Z } };
            DenseMatrix Z = DenseMatrix.OfArray(new double[3, 1] { { v.Position.X }, { v.Position.Y }, { v.Position.Z } });

            DenseMatrix temp = (DenseMatrix)H.Multiply(x);
            DenseMatrix y = (DenseMatrix)Z.Subtract(temp);

            temp = (DenseMatrix)H.Multiply(P);
            DenseMatrix S = (DenseMatrix)temp.Multiply(H.Transpose());
            S = (DenseMatrix)S.Add(R);

            temp = (DenseMatrix)P.Multiply(H.Transpose());
            DenseMatrix K = (DenseMatrix)temp.Multiply(S.Inverse());

            temp = (DenseMatrix)K.Multiply(y);
            x = (DenseMatrix)x.Add(temp);
            temp = (DenseMatrix)K.Multiply(H);
            temp = (DenseMatrix)I.Subtract(temp);
            P = (DenseMatrix)temp.Multiply(P);

            // prediction
            x = (DenseMatrix)F.Multiply(x);
            temp = (DenseMatrix)F.Multiply(P);
            P = (DenseMatrix)temp.Multiply(F.Transpose());

            v.Position.X = (float)x.Values.ElementAt(0);
            v.Position.Y = (float)x.Values.ElementAt(1);
            v.Position.Z = (float)x.Values.ElementAt(2);

            return v;
        }


        /*-------------------------------------------------------------------------
          Display the floor 
          ---------------------------------------------------------------------------*/
        static void displayfloor()
        {
            int i;

            GL.LineWidth(1.0f);
            GL.Begin(BeginMode.Lines);
            GL.Color3(0.5f, 0.5f, 0.5f);    //draw grids

            for (i = 0; i <= 1000; i++)
            {
                GL.Vertex3(i - 500 + centerX, -500 + centerY, centerZ);
                GL.Vertex3(i - 500 + centerX, 500 + centerY, centerZ);
            }

            GL.Color3(0.5f, 0.5f, 0.5f);   //conversion from 1 -> 255
            
            for (i = 0; i <= 1000; i++)
            {
                GL.Vertex3(-500 + centerX, i - 500 + centerY, centerZ);
                GL.Vertex3(500 + centerX, i - 500 + centerY, centerZ);
            }
            
            GL.End();
        }


        /*-------------------------------------------------------------------------
        Compute the center of the model
        ---------------------------------------------------------------------------*/
        static void center()
        {
            //compute the coordinates of the boundary
            for (int i = 0; i < vertices.Count; i++)
            {
                Vertex v = vertices[i];
                centerX += v.Position.X;
                centerY += v.Position.Y;
                centerZ += v.Position.Z;
            }

            centerX = centerX / vertices.Count;
            centerY = centerY / vertices.Count;
            centerZ = centerZ / vertices.Count;
        }



        /*---------------------------------------------------------------------------
        Display the x,y,z axes
        --------------------------------------------------------------------------*/
        static void displayaxes()
        {
            GL.LineWidth(5.0f);
            GL.Begin(BeginMode.Lines);

            GL.Color3(1.0f, 0.0f, 0.0f);  //conversion from 1 -> 255

            GL.Vertex3(centerX, centerY, centerZ);
            GL.Vertex3(1 + centerX, centerY, centerZ);

            GL.Color3(0.0, 1.0f, 0.0);  //conversion from 1 -> 255
            GL.Vertex3(centerX, centerY, centerZ);
            GL.Vertex3(centerX, 1 + centerY, centerZ);

            GL.Color3(1.0f, 1.0f, 0.0);  //conversion from 1 -> 255
            GL.Vertex3(centerX, centerY, centerZ);
            GL.Vertex3(centerX, centerY, 1 + centerZ);
            GL.End();
            GL.LineWidth(1.0f);
        }



        /*---------------------------------------------------------------------------
           Comput and Draw the  bounding box
           --------------------------------------------------------------------------*/
        static void drawboundingbox()
        {
            double max_x = -1e100, max_y = -1e100, max_z = -1e100;
            double min_x = 1e100, min_y = 1e100, min_z = 1e100;

            //compute the coordinates of the boundary
            for (int i = 0; i < vertices.Count; i++)
            {
                Vertex v = vertices[i];
                if (v.Position.X < min_x)
                    min_x = v.Position.X;
                if (v.Position.Y < min_y)
                    min_y = v.Position.Y;
                if (v.Position.Z < min_z)
                    min_z = v.Position.Z;

                if (v.Position.X > max_x)
                    max_x = v.Position.X;
                if (v.Position.Y > max_y)
                    max_y = v.Position.Y;
                if (v.Position.Z > max_z)
                    max_z = v.Position.Z;
            }

            GL.Begin(BeginMode.LineStrip);


            GL.Color3(0, 0, 255);               //conversion from 1 -> 255
            GL.Vertex3(min_x, min_y, min_z);
            GL.Vertex3(min_x, max_y, min_z);
            GL.Vertex3(max_x, max_y, min_z);
            GL.Vertex3(max_x, min_y, min_z);

            GL.Vertex3(min_x, min_y, min_z);

            GL.Vertex3(min_x, min_y, max_z);
            GL.Vertex3(min_x, max_y, max_z);
            GL.Vertex3(max_x, max_y, max_z);
            GL.Vertex3(max_x, min_y, max_z);

            GL.Vertex3(min_x, min_y, max_z);

            GL.End();


            GL.Begin(BeginMode.Lines);
            GL.Color3(0, 0, 255);               //conversion from 1 -> 255
            GL.Vertex3(max_x, max_y, min_z);
            GL.Vertex3(max_x, max_y, max_z);

            GL.Vertex3(min_x, max_y, min_z);
            GL.Vertex3(min_x, max_y, max_z);

            GL.Vertex3(max_x, min_y, min_z);
            GL.Vertex3(max_x, min_y, max_z);
            GL.End();
        }

        #endregion

        //Take care of the Key-presses
        private void MyGlControl_KeyDown(object sender, KeyEventArgs e)
        {

            if (e.KeyCode == Keys.Escape)
                this.Close();

            else if (e.KeyCode == Keys.P)
            {
                current_object_mode = (int)Modes.OBJ_POINT;
                RdPointMode.Checked = true;
            }

            else if (e.KeyCode == Keys.L)
            {
                current_object_mode = (int)Modes.OBJ_LINE;
                RdLineMode.Checked = true;
            }

            else if (e.KeyCode == Keys.D1)
            {
                //LoadModel();
                center();
            }

            else if (e.KeyCode == Keys.B && draw_bounding_box == true) draw_bounding_box = false;
            else if (e.KeyCode == Keys.B && draw_bounding_box == false) draw_bounding_box = true;

            else if (e.KeyCode == Keys.A && draw_axis == true) draw_axis = false;
            else if (e.KeyCode == Keys.A && draw_axis == false) draw_axis = true;

            else if (e.KeyCode == Keys.X && draw_xy_plane == true) draw_xy_plane = false;
            else if (e.KeyCode == Keys.X && draw_xy_plane == false) draw_xy_plane = true;

            // clear vertices
            else if (e.KeyCode == Keys.C)
            {
                vertices.Clear();
                writeToFile.Close();
                writeToFile = new StreamWriter(FileWriteVertices, false);
                ghistResultConfidence.Clear();
                ghistResultIndex.Clear();
                //lscResultConfidence.Clear();
                //lscResultIndex.Clear();
            }

            // check different angles
            else if (e.KeyCode == Keys.Q)
            {
                RdFrontView.Checked = true;  //This command is going all the trick
            }

            // check different angles
            else if (e.KeyCode == Keys.W)
            {
                RdSideView.Checked = true;  ////This command is going all the trick
            }

        }

        //Take care of mouse movement for - rotating, translating, enlarging or shrinking
        private void MyGlControl_MouseMove(object sender, MouseEventArgs e)
        {

            if (e.Button == System.Windows.Forms.MouseButtons.Left)
            {
                xform_mode = (int)Transform.TRANSFORM_ROTATE;
            }
            else if (e.Button == System.Windows.Forms.MouseButtons.Middle)
            {
                xform_mode = (int)Transform.TRANSFORM_TRANSLATE;
            }
            else if (e.Button == System.Windows.Forms.MouseButtons.Right)
            {
                // press_x = Mouse.X; press_y = Mouse.Y;
                xform_mode = (int)Transform.TRANSFORM_SCALE;
            }
            else
            {
                xform_mode = (int)Transform.TRANSFORM_NONE;
            }

            //My Movement when some mouse click is on
            mymotion(e.X, e.Y);
        }


        #region Mouse Motion

        /*---------------------------------------------------------------------------
          Mouse movement callback routine
          --------------------------------------------------------------------------*/

        void mymotion(int x, int y)
        {

            //Console.WriteLine(x + " " + y + " "  + delta);

            if (xform_mode == (int)Transform.TRANSFORM_ROTATE)
            {
                x_angle += (x - press_x) / 3.0f;

                if (x_angle > 180)
                    x_angle -= 360;
                else if (x_angle < -180)
                    x_angle += 360;

                //press_x = x;

                y_angle += (y - press_y) / 3.0f;

                if (y_angle > 180)
                    y_angle -= 360;
                else if (y_angle < -180)
                    y_angle += 360;

                //press_y = y;

                z_angle = (x_angle + y_angle) / 2;
                if (z_angle > 180)
                    z_angle -= 360;
                else if (y_angle < -180)
                    z_angle += 360;


            }
            else if (xform_mode == (int)Transform.TRANSFORM_TRANSLATE)
            {
                x_size += (x - press_x) / 60.0f;
                //press_x = x;

                y_size += (press_y - y) / 60.0f;
                //press_y = y;
            }

            else if (xform_mode == (int)Transform.TRANSFORM_SCALE)
            {
                float old_size = scale_size;

                //scale_size *= (1 + (y - press_y) / 60.0f);

                scale_size *= (1 + ((x - press_x) + (press_y - y)) / 15.0f);


                if (scale_size < 0)
                    scale_size = old_size;
                //press_y = y;
            }

            // force the redraw function
            //glutPostRedisplay();

            //Console.WriteLine(x_size + " " + y_size);

            press_x = x;
            press_y = y;

        }//mymotion (mouse)

        #endregion


        //This method  is called every time the skeleton data is ready to be used.
        private void AllFramesReady(object sender, AllFramesReadyEventArgs e)
        {
            Skeleton[] skeletons = new Skeleton[0];
            DepthImageStream depthStream = sensor.DepthStream;

            using (SkeletonFrame skeletonFrame = e.OpenSkeletonFrame())
            {
                using (DepthImageFrame depthFrame = e.OpenDepthImageFrame())
                {

                    if (skeletonFrame != null && depthFrame != null)
                    {

                        skeletons = new Skeleton[skeletonFrame.SkeletonArrayLength];
                        skeletonFrame.CopySkeletonDataTo(skeletons);

                        depthFrame.CopyDepthImagePixelDataTo(_DepthPixels);

                        if (skeletons.Length != 0)
                        {

                            int counter_skeleton = 0;
                            for (int _ske = 0; _ske < skeletons.Length; _ske++)
                            {
                                if (skeletons[_ske].TrackingState == SkeletonTrackingState.Tracked)
                                    counter_skeleton++;
                            }//for all skeletons


                            foreach (Skeleton skel in skeletons)
                            {

                                if (skel.TrackingState == SkeletonTrackingState.Tracked)
                                {
                                    //Good data here
                                    //This skeleton is good


                                    Joint right_hand = skel.Joints[JointType.HandRight];
                                    Joint right_elbow = skel.Joints[JointType.ElbowRight];

                                    if (IsRightHandDrawing == false)
                                    {
                                        right_hand = skel.Joints[JointType.HandLeft];
                                        right_elbow = skel.Joints[JointType.ElbowLeft];
                                    }

                                    if (right_hand.TrackingState == JointTrackingState.Tracked)
                                    {

                                        //Let's see if the joint information is accurate or inferred
                                        x_right_hand = SkeletonPointToScreen_X(right_hand.Position);
                                        y_right_hand = SkeletonPointToScreen_Y(right_hand.Position);

                                        depth_right_hand = 0;


                                        pixelIndex = x_right_hand + y_right_hand * WIDTH_IMAGE;
                                        if (pixelIndex < WIDTH_IMAGE * HEIGHT_IMAGE)
                                        {
                                            depth_right_hand = _DepthPixels[pixelIndex].Depth;
                                        }


                                        //Looking for the Hand "ABOVE" Elbow Gesture
                                        //If you get X such skeleton frames in which hand is above elbow, start drawing
                                        if (IsPaused == false && IsRegisteringHandMovement == false && right_elbow.TrackingState == JointTrackingState.Tracked &&
                                           right_hand.Position.Y > right_elbow.Position.Y + SAFE_DISTANCE_BW_HAND_ELBOW_ABOVE)
                                        {

                                            CurrentVerticesEstablishedAboveElbow++;

                                            if (CurrentVerticesEstablishedAboveElbow >= TotalNumberOfVerticesToEstablishHandStableAboveElbow)
                                            {
                                                IsRegisteringHandMovement = true;
                                                CurrentVerticesEstablishedAboveElbow = 0;
                                            }


                                        }


                                        //Looking for the Hand "BELOW" Elbow Gesture
                                        //If you get X such skeleton frames in which hand is below elbow, reset the whole drawing
                                        if (IsRegisteringHandMovement == true && right_elbow.TrackingState == JointTrackingState.Tracked &&
                                           right_hand.Position.Y < right_elbow.Position.Y - SAFE_DISTANCE_BW_HAND_ELBOW_BELOW)
                                        {
                                            CurrentVerticesEstablishedBelowElbow++;

                                            if (CurrentVerticesEstablishedBelowElbow >= TotalNumberOfVerticesToEstablishReset)
                                            {

                                                #region Resetting - Hand Tracking Off

                                                IsRegisteringHandMovement = false;
                                                vertices.Clear();
                                                writeToFile.Close();
                                                writeToFile = new StreamWriter(FileWriteVertices, false);
                                                ghistResultConfidence.Clear();
                                                ghistResultIndex.Clear();
                                                //lscResultConfidence.Clear();
                                                //lscResultIndex.Clear();

                                                NUMBER_OF_IMAGES = 10;
                                                TxtTotalResults.Visible = false;

                                                #endregion

                                                //MessageBox.Show("Reset");

                                                CurrentVerticesEstablishedBelowElbow = 0;
                                            }

                                        }

                                        else
                                        {
                                            CurrentVerticesEstablishedBelowElbow = 0;
                                        }




                                        if (depth_right_hand > 0 && IsRegisteringHandMovement == true)
                                        {

                                            /* UNCOMMENT THIS TO ENABLE THRESHOLDING (PART 1 of 2)
                                            if (vertices.Count > 0)
                                            {
                                                float x1 = vertices.ElementAt(vertices.Count - 1).Position.X;
                                                float y1 = vertices.ElementAt(vertices.Count - 1).Position.Y;
                                                float z1 = vertices.ElementAt(vertices.Count - 1).Position.Z;


                                                float x2 = (float)x_right_hand / WIDTH_IMAGE;
                                                float y2 = 1.0f - (float)y_right_hand / HEIGHT_IMAGE;
                                                float z2 = (float)depth_right_hand / NORMALIZE_DEPTH;

                                                //float distance = (x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1) + (z2 - z1) * (z2 - z1); //L2

                                                float distance = Math.Abs(x2 - x1) + Math.Abs(y2 - y1) + Math.Abs(z2 - z1); //L1

                                                if (distance < THRESHOLD_MINIMUM_DISTANCE_TO_QUALIFY_LONG_DISTANCE_ERROR)
                                                {

                                                    int total_intermediate_points = (int)(distance / THRESHOLD_POINTS_INTERPOLATION) - 1;

                                                    //if (total_intermediate_points > 0)
                                                    //richTextBox1.Text += "p";


                                                    for (int _current_inter_point = 1; _current_inter_point <= total_intermediate_points; _current_inter_point++)
                                                    {

                                                        float x3 = (x1 * _current_inter_point) + (x2 * (total_intermediate_points + 1 - _current_inter_point));
                                                        x3 /= (total_intermediate_points + 1);

                                                        float y3 = (y1 * _current_inter_point) + (y2 * (total_intermediate_points + 1 - _current_inter_point));
                                                        y3 /= (total_intermediate_points + 1);

                                                        float z3 = (z1 * _current_inter_point) + (z2 * (total_intermediate_points + 1 - _current_inter_point));
                                                        z3 /= (total_intermediate_points + 1);

                                                        //Add the intermediate (interpolated) point 
                                                        Vertex vertex;
                                                        vertex.Position.X = x3;
                                                        vertex.Position.Y = y3;
                                                        vertex.Position.Z = z3;
                                                        vertex.Normal.X = 0; vertex.Normal.Y = 0; vertex.Normal.Z = 0;
                                                        vertices.Add(vertex);

                                                        writeToFile.WriteLine("v " + vertex.Position.X.ToString() + " " + vertex.Position.Y + " " + vertex.Position.Z.ToString());

                                                    }//for all the intermediate points

                                                }//if this distance is not very very large

                                                    #region Add Vertex
                                                    Vertex original_vertex = new Vertex();
                                                    original_vertex.Position.X = x2;
                                                    original_vertex.Position.Y = y2;
                                                    original_vertex.Position.Z = z2;

                                                    original_vertex.Normal.X = 0; original_vertex.Normal.Y = 0; original_vertex.Normal.Z = 0;
                                                    vertices.Add(original_vertex);
                                                    //Console.WriteLine("..Point Added..");

                                                    writeToFile.WriteLine("v " + original_vertex.Position.X.ToString() + " " + original_vertex.Position.Y + " " + original_vertex.Position.Z.ToString());

                                                    // Console.WriteLine(depth_right_hand);
                                                    #endregion

                                            }//if there are more than 0 vertices in the vertices object

                                                
                                            else if (vertices.Count == 0)
                                            {
                                               */
                                            //UNCOMMENT THIS TO ENABLE THRESHOLDING (PART 1 of 2) 


                                            #region Add Vertex

                                            Vertex original_vertex = new Vertex();
                                            original_vertex.Position.X = (float)x_right_hand / WIDTH_IMAGE;
                                            original_vertex.Position.Y = 1.0f - (float)y_right_hand / HEIGHT_IMAGE;
                                            original_vertex.Position.Z = (float)depth_right_hand / NORMALIZE_DEPTH;

                                            // perform Kalman filter
                                            Vertex smoothed_vertex = performKalman(original_vertex);
                                            smoothed_vertex.Normal.X = 0;
                                            smoothed_vertex.Normal.Y = 0;
                                            smoothed_vertex.Normal.Z = 0;


                                            original_vertex.Normal.X = 0; original_vertex.Normal.Y = 0; original_vertex.Normal.Z = 0;

                                            //vertices.Add(original_vertex);
                                            vertices.Add(smoothed_vertex);
                                            //Console.WriteLine("..Point Added..");

                                            try
                                            {
                                                writeToFile.WriteLine("v " + original_vertex.Position.X.ToString() + " " + original_vertex.Position.Y + " " + original_vertex.Position.Z.ToString());
                                            }
                                            catch { }
                                            // Console.WriteLine(depth_right_hand);

                                            #endregion

                                            /* UNCOMMENT THIS TO ENABLE THRESHOLDING (PART 2 of 2)

                                            }//if this is the first vertex, don't interpolate

                                             */
                                            //UNCOMMENT THIS TO ENABLE THRESHOLDING (PART 2 of 2)

                                        }//if this is a valid depth


                                        //Folow the User with a Big Green Dot, only show nothing saved
                                        if (depth_right_hand > 0 && IsRegisteringHandMovement == false)
                                        {

                                            GL.Color3(0.0f, 1.0f, 0.0f);   //conversion from 1 -> 255
                                            GL.PointSize(10f);
                                            GL.Begin(BeginMode.Points);

                                            GL.Vertex3((float)x_right_hand / WIDTH_IMAGE,
                                                                1.0f - (float)y_right_hand / HEIGHT_IMAGE,
                                                                (float)depth_right_hand / NORMALIZE_DEPTH);

                                            GL.End();


                                        }//if its not the time to save vertices


                                    }//if the hand joint is tracked


                                }//if this is one of the tracked skeleton

                                else if (skel.TrackingState == SkeletonTrackingState.PositionOnly)
                                {
                                    //No use for us
                                }

                            }//for all the skeletons



                        }//if there is at least one skeleton, tracked, untracked, ..


                    }//if both skeleton and frame data exists



                }//depth
            }//skeleton


        }////event AllFramesReady



        #region Kinect Helper Functions

        static private int SkeletonPointToScreen_X(SkeletonPoint skelpoint)
        {
            // Convert point to depth space - X.  
            DepthImagePoint depthPoint = sensor.CoordinateMapper.MapSkeletonPointToDepthPoint(skelpoint, DepthImageFormat.Resolution640x480Fps30);
            return depthPoint.X;
        }
        static private int SkeletonPointToScreen_Y(SkeletonPoint skelpoint)
        {
            // Convert point to depth space - Y.  
            DepthImagePoint depthPoint = sensor.CoordinateMapper.MapSkeletonPointToDepthPoint(skelpoint, DepthImageFormat.Resolution640x480Fps30);
            return depthPoint.Y;
        }

        #endregion


        //Function to make sure ...Sound from kinect is found..
        private static RecognizerInfo GetKinectRecognizer()
        {
            foreach (RecognizerInfo recognizer in SpeechRecognitionEngine.InstalledRecognizers())
            {
                string value;
                recognizer.AdditionalInfo.TryGetValue("Kinect", out value);
                if ("True".Equals(value, StringComparison.OrdinalIgnoreCase) && "en-US".Equals(recognizer.Culture.Name, StringComparison.OrdinalIgnoreCase))
                {
                    return recognizer;
                }
            }

            return null;
        }


        //When a phrase is spoken that is included in the program's Grammer, this function is called.
        private void SpeechRecognized(object sender, SpeechRecognizedEventArgs e)
        {

            //This Confidence number goes from 0 to 1. If you set this value high, the program will be more strict in accepting matches
            //You have to play around with this number for different enviroments.
            const double ConfidenceThreshold = 0.4;

            if (e.Result.Confidence >= ConfidenceThreshold && IsHearingOn == true)
            {

                TxtVoiceCommand.Text = e.Result.Semantics.Value.ToString();

                if (TxtVoiceCommand.Text == "Right")
                {
                    PicRightHand_Click(sender, new EventArgs());
                }

                else if (TxtVoiceCommand.Text == "Left")
                {
                    PicLeftHand_Click(sender, new EventArgs());
                }

                else if (TxtVoiceCommand.Text == "Switch")
                {
                    if (IsRightHandDrawing == true)
                        PicLeftHand_Click(sender, new EventArgs());

                    else if (IsRightHandDrawing == false)
                        PicRightHand_Click(sender, new EventArgs());
                }

                else if (TxtVoiceCommand.Text == "Reset")
                {
                    IsRegisteringHandMovement = true;
                    vertices.Clear();
                    writeToFile.Close();
                    writeToFile = new StreamWriter(FileWriteVertices, false);
                    ghistResultConfidence.Clear();
                    ghistResultIndex.Clear();
                    //lscResultConfidence.Clear();
                    //lscResultIndex.Clear();
                }

                else if (TxtVoiceCommand.Text == "Start")
                {
                    if (IsPaused == true && PicResults == null)
                    {
                        IsRegisteringHandMovement = true;
                        IsPaused = false;
                    }

                    // make them say restart to start from scratch if it's paused
                    // otherwise resume from the current drawing
                    else if (IsPaused)
                    {
                        IsRegisteringHandMovement = true;
                        IsPaused = false;
                    }

                    else  //Real Start Method
                    {
                        IsRegisteringHandMovement = true;
                        vertices.Clear();
                        writeToFile.Close();
                        writeToFile = new StreamWriter(FileWriteVertices, false);
                        ghistResultConfidence.Clear();
                        ghistResultIndex.Clear();
                        //lscResultConfidence.Clear();
                        //lscResultIndex.Clear();
                    }
                }


                else if (TxtVoiceCommand.Text == "Exit")
                {
                    this.Close();
                }

                else if (TxtVoiceCommand.Text == "Perspective")
                {
                    x_angle = 0.0f;
                    y_angle = 0.0f;
                    z_angle = 0.0f;
                    scale_size = 2.0f;
                    x_size = -0.5f;
                    y_size = -0.4f;
                    RdFrontView.Checked = true;
                }

                else if (TxtVoiceCommand.Text == "SideView")
                {
                    x_angle = -80.0f;
                    y_angle = 0.0f;
                    z_angle = 0.0f;
                    scale_size = 2.0f;
                    RdSideView.Checked = true;
                }

                else if (TxtVoiceCommand.Text == "PointMode")
                {
                    current_object_mode = (int)Modes.OBJ_POINT;
                    RdPointMode.Checked = true;
                }

                else if (TxtVoiceCommand.Text == "LineMode")
                {
                    current_object_mode = (int)Modes.OBJ_LINE;
                    RdLineMode.Checked = true;
                }

                else if (TxtVoiceCommand.Text == "Search")
                {
                    BtnDone_Click(sender, new EventArgs());
                }


                else if (TxtVoiceCommand.Text == "ShowMoreResults")
                {
                    BtnShowMoreResults_Click(sender, new EventArgs());
                }


                else if (TxtVoiceCommand.Text == "Pause")
                {
                    BtnPause_Click(sender, new EventArgs());
                }

                else if (TxtVoiceCommand.Text == "Save")
                {
                    saveDraw_Click(sender, new EventArgs());
                }

            }//we are confident abt this result

        }//event SpeechRecognized

        private void SpeechRejected(object sender, SpeechRecognitionRejectedEventArgs e)
        {
            //Voice Cannot be apprehend
        }


        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            writeToFile.Close();

            if (sensor != null)
            {
                sensor.AudioSource.Stop();
                sensor.Stop();
                sensor = null;
            }

            if (speechEngine != null)
            {
                speechEngine.SpeechRecognized -= SpeechRecognized;
                speechEngine.SpeechRecognitionRejected -= SpeechRejected;
                speechEngine.RecognizeAsyncStop();
            }

        }//Form Closing


        #region Hand Selection

        private void PicRightHand_Click(object sender, EventArgs e)
        {
            IsRightHandDrawing = true;
            PicRightHand.Image = global::Final_Demo.Properties.Resources.RightFull;
            PicLeftHand.Image = global::Final_Demo.Properties.Resources.LeftEmpty;
        }

        private void PicLeftHand_Click(object sender, EventArgs e)
        {
            IsRightHandDrawing = false;
            PicLeftHand.Image = global::Final_Demo.Properties.Resources.LeftFull;
            PicRightHand.Image = global::Final_Demo.Properties.Resources.RightEmpty;

        }

        #endregion


        void ShowResults(int NUMBER_OF_IMAGES, List<int> modelIDsA)
        {

            int FIRST_X = 15; //starting from 0, 0 of panel
            int FIRST_Y = 30;

            int Y_MARGIN = 40;

            #region Make Dynamic PictureBoxes and Labels

            MyPanelForImages.Controls.Clear();

            PicResults = new PictureBox[NUMBER_OF_IMAGES];
            TxtConfindence = new Label[NUMBER_OF_IMAGES];

            Font myFont = new System.Drawing.Font("Century", 15.0F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));

            if (modelIDsA.Count > 0)
            {
                for (int i = 0; i < NUMBER_OF_IMAGES; i++)
                {
                    PicResults[i] = new PictureBox();

                    PicResults[i].Location = new System.Drawing.Point(FIRST_X + 20, FIRST_Y + (i) * (120 + Y_MARGIN));
                    PicResults[i].Size = new System.Drawing.Size(120, 120);
                    PicResults[i].BorderStyle = BorderStyle.FixedSingle;
                    PicResults[i].Image = Image.FromFile(ThumbNailLocation + "m" + modelIDsA.ElementAt(i).ToString() + "_thumb.jpg");

                    MyPanelForImages.Controls.Add(PicResults[i]);

                    ///////////////////////////////////////////////////////////////////////////////////////////////////////////
                    ///////////////////////////////////////////////////////////////////////////////////////////////////////////

                    TxtConfindence[i] = new Label();

                    TxtConfindence[i].AutoSize = true;
                    TxtConfindence[i].Font = myFont;
                    TxtConfindence[i].Location = new System.Drawing.Point(FIRST_X + 40, FIRST_Y + (i) * (120 + Y_MARGIN) + 120 - 5);
                    TxtConfindence[i].Size = new System.Drawing.Size(112, 33);
                    TxtConfindence[i].Text = (ghistResultConfidence[i]).ToString("n4");
                    TxtConfindence[i].BackColor = System.Drawing.SystemColors.ActiveCaption;

                    MyPanelForImages.Controls.Add(TxtConfindence[i]);

                }//for all the images

            }//if there is at least one address

            PicResults = new PictureBox[NUMBER_OF_IMAGES];
            TxtConfindence = new Label[NUMBER_OF_IMAGES];
			
            #endregion

        }//ShowResults()


        private void PCA_Threshold()
        {
            Process b = new Process();
            b.StartInfo.RedirectStandardOutput = true;
            b.StartInfo.RedirectStandardError = true;
            b.StartInfo.FileName = CppExeLocation;
            b.StartInfo.Arguments = CppArguments;
            Debug.WriteLine(CppArguments);
            Debug.WriteLine(b.StartInfo.Arguments);
            b.StartInfo.UseShellExecute = false;
            b.Start();
            string output = b.StandardOutput.ReadToEnd();
            string error = b.StandardError.ReadToEnd();
            
            ////MessageBox.Show(output);
            ////MessageBox.Show(error);
            //b.WaitForExit();
        }

        // Load a previously drawn picture, and search it.
        private void LoadFile_Click(object sender, EventArgs e)
        {

            OpenFileDialog openFileDialog1 = new OpenFileDialog();
            openFileDialog1.InitialDirectory = Directory.GetParent(Directory.GetParent(Directory.GetCurrentDirectory()).FullName).FullName + "\\Points";
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                writeToFile.Close();
                System.IO.StreamReader sr = new System.IO.StreamReader(openFileDialog1.FileName);
                //writeToFile = new StreamWriter(FileWriteVertices, false);
                
                //while (sr.Peek() != -1) // stops when it reachs the end of the file
                //{
                //    string line = sr.ReadLine();
                //    writeToFile.WriteLine(line);
                //}

                try
                {
                    System.IO.File.Copy(openFileDialog1.FileName, "Vertices.obj", true);
                }
                catch
                {
                    Thread.Sleep(3000);
                    try
                    {
                        System.IO.File.Copy(openFileDialog1.FileName, "Vertices.obj", true);
                    }
                    catch
                    {
                        Thread.Sleep(3000);
                        System.IO.File.Copy(openFileDialog1.FileName, "Vertices.obj", true);
                    }
                }

                //MessageBox.Show(openFileDialog1.FileName);
                //sr.Close();

                #region Check if necessary files are present.

                if (!Directory.Exists(GHIST_DIR))
                {
                    MessageBox.Show("Sorry, but the file \"" + Directory.GetCurrentDirectory() + GHIST_DIR + "\" was not found!");
                    return;
                }

                if (!File.Exists(GHIST_JAR))
                {
                    MessageBox.Show("Sorry, but the file \"" + Directory.GetCurrentDirectory() + GHIST_JAR + "\" was not found!");
                    return;
                }

                //if (!Directory.Exists(LSC_DIR))
                //{
                //    MessageBox.Show("Sorry, but the file \"" + Directory.GetCurrentDirectory() + LSC_DIR + "\" was not found!");
                //    return;
                //}

                //if (!File.Exists(LSC_EXEC))
                //{
                //    MessageBox.Show("Sorry, but the file \"" + Directory.GetCurrentDirectory() + LSC_EXEC + "\" was not found!");
                //    return;
                //}

                if (!Directory.Exists("Thumbnails"))
                {
                    MessageBox.Show("Sorry, but the folder \"" + Directory.GetCurrentDirectory() + "\\Thumbnails\" was not found!");
                    return;
                }


                #endregion

                // Run the program(s) on the sketch to get result file of distances between other models.
                PCA_Threshold();

                // Create process to run global histogram code and start it.
                // Integrated by Mario on 7/28/2014 at 3:00 pm
                Process ghist = new Process();
                ghist.StartInfo.FileName = JAVA_EXEC;
                ghist.StartInfo.Arguments = GHIST_ARG;
                ghist.StartInfo.UseShellExecute = false;
                ghist.Start();
                ghist.WaitForExit();

                // Create process to run LSC code and start it.
                //Process lsc = new Process();
                //lsc.StartInfo.FileName = LSC_EXEC;
                //lsc.StartInfo.Arguments = LSC_ARG;
                //lsc.Start();
                //lsc.WaitForExit();

                System.Threading.Thread.Sleep(2000);

                #region Read the Results

                //StreamReader ReadResults = new StreamReader(FileFinalResults);
                StreamReader ghistResults = new StreamReader(GHIST_OUT);
                //StreamReader lscResults = new StreamReader(LSC_OUT);
                string ReadLine;
                string[] bits;
                float Confidence;
                int Index;

                // Read in GHIST results.
                while (true)
                {
                    ReadLine = ghistResults.ReadLine();

                    if (ReadLine == null || ReadLine == "")
                        break;

                    bits = ReadLine.Split(' ');
                    Confidence = float.Parse(bits[0]);
                    Index = Int32.Parse(bits[1]);

                    ghistResultConfidence.Add(Confidence);
                    ghistResultIndex.Add(Index);

                } // while reading results

                ghistResults.Close();

                // Read in LSC results.
                //while (true)
                //{
                //    ReadLine = lscResults.ReadLine();

                //    if (ReadLine == null || ReadLine == "")
                //        break;

                //    bits = ReadLine.Split(new char[] { ' ', '\t' });
                //    Confidence = float.Parse(bits[0]);
                //    Index = Int32.Parse(bits[1]);

                //    //lscResultConfidence.Add(Confidence);
                //    //lscResultIndex.Add(Index);

                //} // while reading results

                //lscResults.Close();

                #endregion


                #region Show the Thumbnails and the Confidence Level of each Match


                ShowResults(NUMBER_OF_IMAGES, ghistResultIndex);


                #endregion

                //if (File.Exists(FileFinalResults))
                //    File.Delete(FileFinalResults);

                //writeToFile = new StreamWriter(FileWriteVertices, false);
                writeToFile.Close();
                return;
            }
        }

        private void saveDraw_Click(object sender, EventArgs e){

            if (direcList.SelectedItem != null)
            {
                if (vertices.Count != 0)
                {
                    // class of what user selected to draw.
                    string selectedItem = direcList.SelectedItem.ToString();

                    // retrieving the path that contains the resultPoints
                    string curr_path = Environment.CurrentDirectory.ToString();
                    curr_path = System.String.Concat(curr_path, "\\resultPoints.txt");

                    // getting new path to save resultPoints in
                    string new_path = Path.GetDirectoryName(Path.GetDirectoryName(Path.GetDirectoryName(curr_path)));
                    new_path = System.String.Concat(new_path, "\\Points\\" + selectedItem + "\\");

                    // saves files by numbers 

                    for (int i = 1; ; ++i)
                    {
                        if (!File.Exists(System.String.Concat(new_path, i + ".txt")))
                        {
                            new_path = System.String.Concat(new_path, i + ".txt");
                            break;
                        }
                    }

                    // copying reults file
                    System.IO.File.Copy(curr_path, new_path);
                    TxtVoiceCommand.Text = "Saved";

                }
                else
                {
                    System.Windows.Forms.MessageBox.Show("No file was saved");
                    return;
                }
            }
            else
            {
                System.Windows.Forms.MessageBox.Show("No file was saved");
                return;
            }
        }

        private void direcList_SelIndex(object sender, EventArgs e){
            
        }

        private void BtnDone_Click(object sender, EventArgs e)
        {
            if (vertices.Count == 0)
            {
                TxtVoiceCommand.Text = "No points.";
                return;
            }

            if (IsRegisteringHandMovement == false) //All ready stopped
            {
                TxtVoiceCommand.Text = "Restart please.";
                return;
            }

            TxtTotalResults.Visible = true;

            IsRegisteringHandMovement = false;
            IsPaused = true;

            //Stop Wrtiting to File
            writeToFile.Close();

            // make a copy of the points to be used at a later time
            System.IO.File.Copy("Vertices.obj", "resultPoints.txt", true);

            // Apply PCA AND Uniform Point Sampling on Vertices.obj. Product will be in a text file ThreshVertices.obj


            #region Check if necessary files are present.

            if (!Directory.Exists(GHIST_DIR))
            {
                MessageBox.Show("Sorry, but the file \"" + Directory.GetCurrentDirectory() + GHIST_DIR + "\" was not found!");
                return;
            }

            if (!File.Exists(GHIST_JAR))
            {
                MessageBox.Show("Sorry, but the file \"" + Directory.GetCurrentDirectory() + GHIST_JAR + "\" was not found!");
                return;
            }

            //if (!Directory.Exists(LSC_DIR))
            //{
            //    MessageBox.Show("Sorry, but the file \"" + Directory.GetCurrentDirectory() + LSC_DIR + "\" was not found!");
            //    return;
            //}

            //if (!File.Exists(LSC_EXEC))
            //{
            //    MessageBox.Show("Sorry, but the file \"" + Directory.GetCurrentDirectory() + LSC_EXEC + "\" was not found!");
            //    return;
            //}

            if (!Directory.Exists("Thumbnails"))
            {
                MessageBox.Show("Sorry, but the folder \"" + Directory.GetCurrentDirectory() + "\\Thumbnails\" was not found!");
                return;
            }


            #endregion

            // Run the program(s) on the sketch to get result file of distances between other models.

            // Commented out old code for running .exe file - Mario
            //Process b = new Process();
            //b.StartInfo.RedirectStandardOutput = true;
            //b.StartInfo.RedirectStandardError = true;
            //b.StartInfo.FileName = CppExeLocation;
            //b.StartInfo.Arguments = CppArguments;
            //Debug.WriteLine(CppArguments);
            //Debug.WriteLine(b.StartInfo.Arguments);
            //b.StartInfo.UseShellExecute = false;
            //b.Start();
            //string output = b.StandardOutput.ReadToEnd();
            //string error = b.StandardError.ReadToEnd();
            ////MessageBox.Show(output);
            ////MessageBox.Show(error);
            //b.WaitForExit();

            PCA_Threshold();

            // Create process to run global histogram code and start it.
            // Integrated by Mario on 7/28/2014 at 3:15 pm
            Process ghist = new Process();
            ghist.StartInfo.FileName = JAVA_EXEC;
            ghist.StartInfo.Arguments = GHIST_ARG;
            ghist.StartInfo.UseShellExecute = false;
            ghist.Start();
            ghist.WaitForExit();

            // Create process to run LSC code and start it.
            //Process lsc = new Process();
            //lsc.StartInfo.FileName = LSC_EXEC;
            //lsc.StartInfo.Arguments = LSC_ARG;
            //lsc.StartInfo.UseShellExecute = false;
            //lsc.Start();
            //lsc.WaitForExit();

            System.Threading.Thread.Sleep(2000);   //changed by Bo Li, 04282015

            #region Read the Results

            //StreamReader ReadResults = new StreamReader(FileFinalResults);
            StreamReader ghistResults = new StreamReader(GHIST_OUT);
            //StreamReader lscResults = new StreamReader(LSC_OUT);
            string ReadLine;
            string[] bits;
            float Confidence;
            int Index;

            // Read in GHIST results.
            while (true)
            {
                ReadLine = ghistResults.ReadLine();

                if (ReadLine == null || ReadLine == "")
                    break;

                bits = ReadLine.Split(' ');
                Confidence = float.Parse(bits[0]);
                Index = Int32.Parse(bits[1]);

                ghistResultConfidence.Add(Confidence);
                ghistResultIndex.Add(Index);

            } // while reading results

            ghistResults.Close();

            //// Read in LSC results.
            //while (true)
            //{
            //    ReadLine = lscResults.ReadLine();

            //    if (ReadLine == null || ReadLine == "")
            //        break;

            //    bits = ReadLine.Split(new char[] { ' ', '\t' });
            //    Confidence = float.Parse(bits[0]);
            //    Index = Int32.Parse(bits[1]);

            //    //lscResultConfidence.Add(Confidence);
            //    //lscResultIndex.Add(Index);

            //} // while reading results

            //lscResults.Close();

            #endregion


            #region Show the Thumbnails and the Confidence Level of each Match


            ShowResults(NUMBER_OF_IMAGES, ghistResultIndex);


            #endregion

            //if (File.Exists(FileFinalResults))
            //    File.Delete(FileFinalResults);

            writeToFile = new StreamWriter(FileWriteVertices, false);

        }//BtnDone pressed


        #region Additional GUI

        private void RdFrontView_CheckedChanged(object sender, EventArgs e)
        {
            if (RdFrontView.Checked == true)
            {
                x_angle = 0.0f;
                y_angle = 0.0f;
                z_angle = 0.0f;
                scale_size = 2.0f;
                x_size = -0.5f;
                y_size = -0.4f;
            }

            else
            {
                x_angle = -80.0f;
                y_angle = 0.0f;
                z_angle = 0.0f;
                scale_size = 2.0f;
            }


        }

        private void RdPointMode_CheckedChanged(object sender, EventArgs e)
        {

            if (RdPointMode.Checked == true)
            {
                current_object_mode = (int)Modes.OBJ_POINT;
            }
            else
            {
                current_object_mode = (int)Modes.OBJ_LINE;
            }

        }



        #endregion

        private void BtnSearchMagnify_Click(object sender, EventArgs e)
        {
            BtnDone_Click(sender, e);
        }

        private void PicVoiceOnOff_Click(object sender, EventArgs e)
        {
            if (IsHearingOn == true)
            {
                IsHearingOn = false;
                TxtVoiceCommand.Text = "Voice Off";
                PicVoiceOnOff.Image = global::Final_Demo.Properties.Resources.cross;
            }

            else if (IsHearingOn == false)
            {
                IsHearingOn = true;
                TxtVoiceCommand.Text = "Listening...";
                PicVoiceOnOff.Image = global::Final_Demo.Properties.Resources.tick;
            }


        }

        private void BtnShowMoreResults_Click(object sender, EventArgs e)
        {
            if (PicResults == null)
                return;

            NUMBER_OF_IMAGES += 10;
            TxtTotalResults.Text = "First " + NUMBER_OF_IMAGES.ToString();
            ShowResults(NUMBER_OF_IMAGES, ghistResultIndex);
        }

        private void BtnRestart_Click(object sender, EventArgs e)
        {
            IsRegisteringHandMovement = true;
            vertices.Clear();
            writeToFile.Close();
            writeToFile = new StreamWriter(FileWriteVertices, false);
            ghistResultConfidence.Clear();
            ghistResultIndex.Clear();
            //lscResultConfidence.Clear();
            //lscResultIndex.Clear();

            NUMBER_OF_IMAGES = 10;
            TxtTotalResults.Visible = false;

        }

        private void BtnPause_Click(object sender, EventArgs e)
        {
            if (IsRegisteringHandMovement == true)
            {
                IsRegisteringHandMovement = false;
                IsPaused = true;
            }
            else
            {
                IsRegisteringHandMovement = true;
                IsPaused = false;
            }
        }


    }//class
}//namespace