package edu.txstate.mag262;
import java.util.ArrayList;

/** A bounding sphere of a set of points, having radius and center. */
public class BoundingSphere {
	
	private double radius;
	private Vertex3D center;
	
	private BoundingSphere() {}
	
	/** Called internally, instantiates a bounding sphere object. */
	private BoundingSphere(double radius, Vertex3D center) {
		this.radius = radius;
		this.center = center;
	}
	
	/** Calculates a bounding sphere for a group of points in
	 *  3D space using the bounding box.
	 *  
	 * @param vertices	the vertices to build the sphere from.
	 */
	public static BoundingSphere boundingBox(ArrayList<Vertex3D> vertices) {
		Vertex3D mp = new Vertex3D(vertices.get(0));
		Vertex3D MP = new Vertex3D(mp);
		for (Vertex3D v : vertices) {
			if (v.x < mp.x) mp.x = v.x;
			if (v.x > MP.x) MP.x = v.x;
			if (v.y < mp.y) mp.y = v.y;
			if (v.y > MP.y) MP.y = v.y;
			if (v.z < mp.z) mp.z = v.z;
			if (v.z > MP.z) MP.z = v.z;
		}
		double radius = mp.distanceTo(MP) / 2.0;
		Vertex3D center = mp.midpoint(MP);
		return new BoundingSphere(radius, center);
	}
	
	/** Shifts points to origin by bounding sphere center.
	 * 
	 * @param points	model vertices.
	 */
	public void toOrigin(ArrayList<Vertex3D> points) {
		for (Vertex3D v : points) v.sub(center);
		center.sub(center);
	}
	
	/** @return The radius of the sphere. */
	public double getRadius() { return radius; }
	
	/** @return The center point of the sphere. 
	 *  @see Vertex3D */
	public Vertex3D getCenter() { return center; }
}
