package edu.txstate.mag262;

import java.util.ArrayList;

/** A useful representation of a class in a Princeton-Shape-Benchmark-style .cla file. */
public class CLAClass {
	
	private String name;
	private String parent;
	private ArrayList<Integer> models;
	
	/** Constructs an instance of this class with the given data.
	 * 
	 * @param name		name of class
	 * @param parent	name of parent class, "0" if no parent
	 * @param models	list of model numbers assigned to this class
	 * @see ArrayList
	 */
	public CLAClass(String name, String parent, ArrayList<Integer> models) {
		this.name = name;
		this.parent = parent;
		this.models = models;
	}
	
	/** Copy constuctor.
	 * 
	 * @param o	other CLAClass object.
	 */
	public CLAClass(CLAClass o) {
		this.name = o.name;
		this.parent = o.parent;
		this.models = new ArrayList<>();
		for (Integer i : o.models)
			this.models.add(i);
	}
	
	/** @return Name of class. */
	public String getName() { return name; }
	
	/** @return Parent-class name, "0" means no parent. */
	public String getParent() { return parent; }
	
	/** @return Pointer to internal list of models. */
	public ArrayList<Integer> getModelList() { return models; }
	
	/** @return Number of models in this class. */
	public int numModels() { return models.size(); }
	
	@Override
	public String toString() {
		String s = "<" + name + "> class";
		if (!parent.equalsIgnoreCase("0"))
				s += " (child of <" + parent + ">)";
		s += "\nIs class of the following " + models.size() + " models:\n";
		for (int i = 0; i < models.size(); ++i)
			s += "\t" + models.get(i) + "\n";
		return s;
	}
}
