package edu.txstate.mag262;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Scanner;

/** Contains entry point of application. */
public class Runner {
	
	private static SphereApproximator sphere = new SphereApproximator(3);
	private static int shells = 2;
	private static double similarity = 2;
	
	public static void main(String[] args) {
		
		// Read in parameters from a file 'params.dat'.
		try (Scanner s = new Scanner(new FileInputStream("params.dat"))) {
			
			// Read in integer values.
			ArrayList<Integer> params = new ArrayList<>();
			while (s.hasNextInt()) params.add(s.nextInt());
			
			// Use polyhedron model.
			if (params.size() == 2) {
				sphere = new SphereApproximator(params.get(0));
				shells = params.get(1);
				System.out.printf("\nORDER(%d), SHELLS(%d)", params.get(0), params.get(1));
				
			// Use spherical coordinates.
			} else if (params.size() == 3) {
				sphere = new SphereApproximator(params.get(0), params.get(1));
				shells = params.get(2);
				System.out.printf("\nH_SEC(%d), V_SEC(%d), SHELLS(%d)", params.get(0), params.get(1), params.get(2));
			}
			
			// Read in similarity matrix parameter.
			if (s.hasNextDouble()) {
				similarity = s.nextDouble();
				System.out.printf(", SIML(%01.1f)\n", similarity);
			} else {
				System.out.println("Malformed parameter file ('params.dat')!");
				System.exit(1);
			}
			
		} catch (FileNotFoundException e) {
			System.out.println("\nParameter file ('params.dat') not found!");
			System.out.println("You better be generating a distance matrix...");
		}
		
		System.out.println();
		
		// Some user output.
		if (args.length == 0) {
			System.out.println("To generate histograms directory:");
			System.out.println("\tjava -jar ghist.jar <MDL_DIR> <OUT_DIR>");
			System.out.println("\tjava -jar ghist.jar <PSB_DIR> <OUT_DIR> --psb");
			System.out.println("\tjava -jar ghist.jar <MDL_DIR> <OUT_DIR> --vanilla\n");
			
			System.out.println("To compare single model against histogram directory:");
			System.out.println("\tjava -jar ghist.jar <HIST_DIR> <(OBJ|OFF)_FILE> <OUT_FILE>");
			System.out.println("\tjava -jar ghist.jar <HIST_DIR> <(OBJ|OFF)_FILE> <OUT_FILE> --vanilla");
			System.out.println("\tjava -jar ghist.jar <HIST_DIR> <OFF_FILE> <MODEL_INFO> <OUT_FILE>\n");
			
			System.out.println("To create distance matrix:");
			System.out.println("\tjava -jar ghist.jar <HIST_DIR> <CLA_FILE> <OUT_FILE>");
			System.out.println("\tjava -jar ghist.jar <HIST_DIR> <CLA_FILE> <OUT_FILE> --generic\n");
		}
		
		// If generating histograms directory, performs normalization.
		else if (args.length == 2) {
			new HistogramDirectory(args[0], args[1], shells, sphere).generate(HistogramDirectory.NORMALIZE);
		}
		
		else if (args.length == 3) {
			
			// If normalizing and comparing a single model against a histogram directory.
			if (args[1].toLowerCase().endsWith("obj") || args[1].toLowerCase().endsWith("off")) {
				new GlobalHistogram(new Model3D(args[1], true), shells, sphere).compareWith(args[0], args[2], similarity);
				
			// If creating distance matrix using .ghist files.
			} else if (args[1].toLowerCase().endsWith("cla")) {
				ShapeDescriptor shape = new ShapeDescriptor(2, shells, sphere);
				GlobalHistogram.createDistanceMatrix(args[0], args[1], args[2], shape, similarity);
			
			// Generating a histogram directory using PSB-style model info files.
			} else if (args[2].toLowerCase().contains("psb")) {
				new HistogramDirectory(args[0], args[1], shells, sphere).generate(HistogramDirectory.MODEL_INFO);
			
			// If generating histograms directory, performing no normalization.
			} else if (args[2].toLowerCase().contains("vanilla")) {
				new HistogramDirectory(args[0], args[1], shells, sphere).generate(HistogramDirectory.VANILLA);
			}
		
		} else if (args.length == 4) {
			
			// If comparing a single model against a histogram directory, using model info for normalization.
			if (args[2].toLowerCase().endsWith("txt") && !args[3].toLowerCase().endsWith("vanilla"))
				new GlobalHistogram(new Model3D(args[1], args[2]), shells, sphere).compareWith(args[0], args[3], similarity);
			
			// If creating distance matrix using straight histogram data, i.e., all bin values separated by white space.
			else if (args[3].toLowerCase().contains("generic")) {
				GlobalHistogram.createDistanceMatrix(args[0], args[1], args[2]);
			
			// If comparing a single model against a histogram directory, without normalization.
			} else if (args[3].toLowerCase().contains("vanilla")) {
				new GlobalHistogram(new Model3D(args[1], false), shells, sphere).compareWith(args[0], args[2], similarity);
			}
		}
	}
}
